//----------------------------------------------------------------------------
//  EDGE Data Definition File Code (Attack Types)
//----------------------------------------------------------------------------
// 
//  Copyright (c) 1999-2000  The EDGE Team.
// 
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//----------------------------------------------------------------------------
//
// Attacks Setup and Parser Code
//
// 1998/10/29 -KM- Finalisation of sound code.  SmartProjectile.
//

#include "i_defs.h"

#include "ddf_locl.h"
#include "ddf_main.h"
#include "dm_state.h"
#include "m_fixed.h"
#include "m_math.h"
#include "p_mobj.h"
#include "z_zone.h"

static attacktype_t buffer_atk;
static attacktype_t *dynamic_atk;

// this (and buffer_mobj) logically belongs with buffer_atk:
static boolean_t attack_has_mobj;
static float_t a_damage_range;
static float_t a_damage_multi;

static const attacktype_t template_atk =
{
  DDF_BASE_NIL,  // ddf

  0,        // attackstyle
  0,        // flags
  NULL,     // initsound
  NULL,     // sound
  0,        // accuracy slope
  0,        // accuracy angle
  0,        // xoffset
  0,        // yoffset
  0,        // angle_offset
  0,        // slope_offset
  0,        // assault_speed
  0,        // height
  0,        // range
  0,        // count
  0,        // tooclose

  // damage info
  {
    0,      // nominal
    -1,     // linear_max
    -1,     // error
    0,      // delay
    NULL_LABEL, NULL_LABEL, NULL_LABEL  // override labels
  },

  BITSET_EMPTY, // attack_class
  0,        // objinitstate
  NULL,     // objinitstate_ref
  PERCENT_MAKE(0), // notracechance
  PERCENT_MAKE(0), // keepfirechance
  NULL,     // atk_mobj
  NULL,     // spawnedobj
  NULL,     // spawnedobj_ref
  NULL,     // puff
  NULL      // puff_ref
};

attacktype_t ** ddf_attacks = NULL;
int num_ddf_attacks = 0;

static stack_array_t ddf_attacks_a;


static void DDF_AtkGetType(const char *info, void *storage, void *priv);
static void DDF_AtkGetSpecial(const char *info, void *storage, void *priv);
static void DDF_AtkGetLabel(const char *info, void *storage, void *priv);

static damage_t dummy_damage;

const commandlist_t damage_commands[] =
{
  {"VAL",   DDF_MainGetFloat,  FIELD_OFF2P(dummy_damage,nominal),  NULL},
  {"MAX",   DDF_MainGetFloat,  FIELD_OFF2P(dummy_damage,linear_max), NULL},
  {"ERROR", DDF_MainGetFloat,  FIELD_OFF2P(dummy_damage,error), NULL},
  {"DELAY", DDF_MainGetTime,   FIELD_OFF2P(dummy_damage,delay), NULL},

  {"PAIN STATE",  DDF_AtkGetLabel, FIELD_OFF2P(dummy_damage,pain), NULL},
  {"DEATH STATE", DDF_AtkGetLabel, FIELD_OFF2P(dummy_damage,death), NULL},
  {"OVERKILL STATE", DDF_AtkGetLabel, FIELD_OFF2P(dummy_damage,overkill), NULL},
  {NULL, NULL, NULL, NULL}
};

// -KM- 1998/09/27 Major changes to sound handling
// -KM- 1998/11/25 Accuracy + Translucency are now fraction.  Added a spare attack for BFG.
// -KM- 1999/01/29 Merged in thing commands, so there is one list of
//  thing commands for all types of things (scenery, items, creatures + projectiles)

static const commandlist_t attack_commands[] =
{
    // sub-commands
    {"*DAMAGE", NULL, &buffer_atk.damage, (void *)damage_commands},

    {"ATTACKTYPE", DDF_AtkGetType, NULL, NULL},
    {"ATTACK SPECIAL", DDF_AtkGetSpecial, NULL, NULL},
    {"ACCURACY SLOPE", DDF_MainGetSlope, &buffer_atk.accuracy_slope, NULL},
    {"ACCURACY ANGLE", DDF_MainGetAngle, &buffer_atk.accuracy_angle, NULL},
    {"ATTACK HEIGHT", DDF_MainGetFloat, &buffer_atk.height, NULL},
    {"SHOTCOUNT", DDF_MainGetNumeric, &buffer_atk.count, NULL},
    {"X OFFSET", DDF_MainGetFloat, &buffer_atk.xoffset, NULL},
    {"Y OFFSET", DDF_MainGetFloat, &buffer_atk.yoffset, NULL},
    {"ANGLE OFFSET", DDF_MainGetAngle, &buffer_atk.angle_offset, NULL},
    {"SLOPE OFFSET", DDF_MainGetSlope, &buffer_atk.slope_offset, NULL},
    {"ATTACKRANGE", DDF_MainGetFloat, &buffer_atk.range, NULL},
    {"TOO CLOSE RANGE", DDF_MainGetNumeric, &buffer_atk.tooclose, NULL},
    {"NO TRACE CHANCE", DDF_MainGetPercent, &buffer_atk.notracechance, NULL},
    {"KEEP FIRING CHANCE", DDF_MainGetPercent, &buffer_atk.keepfirechance, NULL},
    {"ASSAULT SPEED", DDF_MainGetFloat, &buffer_atk.assault_speed, NULL},
    {"ATTEMPT SOUND", DDF_MainLookupSound, &buffer_atk.initsound, NULL},
    {"ENGAGED SOUND", DDF_MainLookupSound, &buffer_atk.sound, NULL},
    {"SPAWNED OBJECT", DDF_MainGetString, &buffer_atk.spawnedobj_ref, NULL},
    {"SPAWN OBJECT STATE", DDF_MainGetString, &buffer_atk.objinitstate_ref, NULL},
    {"PUFF", DDF_MainGetString, &buffer_atk.puff_ref, NULL},
    {"ATTACK CLASS", DDF_MainGetBitSet, &buffer_atk.attack_class, NULL},

    // -AJA- backward compatibility cruft...
    {"!DAMAGE", DDF_MainGetFloat, &buffer_atk.damage.nominal, NULL},
    {"!DAMAGE RANGE", DDF_MainGetFloat, &a_damage_range, NULL},
    {"!DAMAGE MULTI", DDF_MainGetFloat, &a_damage_multi, NULL},
    {NULL, NULL, NULL, NULL}
};


//
//  DDF PARSE ROUTINES
//

static boolean_t AttackStartEntry(const char *name)
{
  int i;
  boolean_t replaces = false;

  if (name && name[0])
  {
    for (i=0; i < num_ddf_attacks; i++)
    {
      if (DDF_CompareName(ddf_attacks[i]->ddf.name, name) == 0)
      {
        dynamic_atk = ddf_attacks[i];
        replaces = true;
        break;
      }
    }
  }

  // not found, create a new one
  if (! replaces)
  {
    Z_SetArraySize(&ddf_attacks_a, ++num_ddf_attacks);
    
    dynamic_atk = ddf_attacks[num_ddf_attacks - 1];
    dynamic_atk->ddf.name = (name && name[0]) ? Z_StrDup(name) :
        DDF_MainCreateUniqueName("UNNAMED_ATTACK", num_ddf_attacks);
  }

  dynamic_atk->ddf.number = 0;

  attack_has_mobj = false;
  a_damage_range = -1;
  a_damage_multi = -1;

  return replaces;
}

static void AttackCopyOld(void)
{
  buffer_atk = dynamic_atk[0];

  // handle attacks that have mobjs
  if (dynamic_atk->atk_mobj)
  {
    buffer_mobj = dynamic_atk->atk_mobj[0];
    attack_has_mobj = true;
  }
  else
    buffer_mobj = template_mobj;
}

static void AttackInstantiate(const char *template_name)
{
  int i;

  if (! template_name)
  {
    // use built-in template
    buffer_atk  = template_atk;
    buffer_mobj = template_mobj;
    return;
  }

  for (i=0; i < num_ddf_attacks; i++)
  {
    // ignore current entry
    if (ddf_attacks[i] == dynamic_atk)
      continue;
    
    if (DDF_CompareName(ddf_attacks[i]->ddf.name, template_name) == 0)
    {
      buffer_atk = ddf_attacks[i][0];

      // handle attacks that have mobjs
      if (dynamic_atk->atk_mobj)
      {
        buffer_mobj = dynamic_atk->atk_mobj[0];
        attack_has_mobj = true;
      }
      else
        buffer_mobj = template_mobj;

      return;
    }
  }

  I_Error("Template `%s' not found for attacks.ddf\n", template_name);
}

static void AttackParseField(const char *field, const char *contents,
    int index, boolean_t is_last)
{
#if (DEBUG_DDF)  
  I_WriteDebug("ATTACK_PARSE: [%s] [%s]\n", field, contents);
#endif

  // first, check attack commands
  if (DDF_MainParseField(attack_commands, field, contents))
    return;

  // we need to create an MOBJ for this attack
  attack_has_mobj = true;

  ThingParseField(field, contents, index, is_last);
}

static void AttackFinishEntry(void)
{
  ddf_base_t base;
  
  // FIXME: check stuff...

  // handle attacks that have mobjs
  if (attack_has_mobj)
  {
    if (buffer_mobj.first_state)
      DDF_StateFinishStates(buffer_mobj.first_state, 
          buffer_mobj.last_state);

    buffer_atk.atk_mobj = DDF_MobjMakeAttackObj(&buffer_mobj,
      dynamic_atk->ddf.name);
  }
  else
    buffer_atk.atk_mobj = NULL;
  
  // compute an attack class, if none specified
  if (buffer_atk.attack_class == BITSET_EMPTY)
  {
    buffer_atk.attack_class = attack_has_mobj ? BITSET_MAKE('M') : 
        (buffer_atk.attackstyle == ATK_CLOSECOMBAT ||
         buffer_atk.attackstyle == ATK_SKULLFLY) ? 
        BITSET_MAKE('C') : BITSET_MAKE('B');
  }

  // -AJA- 2001/01/27: Backwards compatibility
  if (a_damage_range > 0)
  {
    buffer_atk.damage.nominal = a_damage_range;

    if (a_damage_multi > 0)
      buffer_atk.damage.linear_max = a_damage_range * a_damage_multi;
  }

  // transfer static entry to dynamic entry
  
  base = dynamic_atk->ddf;
  dynamic_atk[0] = buffer_atk;
  dynamic_atk->ddf = base;

  // compute CRC...
  CRC32_Init(&dynamic_atk->ddf.crc);

  // FIXME: add more stuff...

  CRC32_Done(&dynamic_atk->ddf.crc);
}


void DDF_ReadAtks(void *data, int size)
{
  readinfo_t attacks;

  attacks.memfile = data;
  attacks.memsize = size;
  attacks.tag = "ATTACKS";
  attacks.entries_per_dot = 2;

  if (attacks.memfile)
  {
    attacks.message = NULL;
    attacks.filename = NULL;
    attacks.lumpname = "DDFATK";
  }
  else
  {
    attacks.message = "DDF_InitAttacks";
    attacks.filename = "attacks.ddf";
    attacks.lumpname = NULL;
  }

  attacks.start_entry  = AttackStartEntry;
  attacks.copy_old     = AttackCopyOld;
  attacks.instantiate  = AttackInstantiate;
  attacks.parse_field  = AttackParseField;
  attacks.finish_entry = AttackFinishEntry;

  DDF_MainReadFile(&attacks);
}

void DDF_AttackInit(void)
{
  Z_InitStackArray(&ddf_attacks_a, (void ***)&ddf_attacks, sizeof(attacktype_t), 0);
}

void DDF_AttackCleanUp(void)
{
  int i;
  attacktype_t *atk;

  for (i=0; i < num_ddf_attacks; i++)
  {
    atk = ddf_attacks[i];
    
    DDF_ErrorSetEntryName("[%s]  (attacks.ddf)", atk->ddf.name);

    // lookup thing references

    atk->puff = atk->puff_ref ? DDF_MobjLookup(atk->puff_ref) : NULL;

    atk->spawnedobj = atk->spawnedobj_ref ? 
        DDF_MobjLookup(atk->spawnedobj_ref) : NULL;
      
    if (atk->spawnedobj)
    {
      if (atk->objinitstate_ref)
        atk->objinitstate = DDF_MainLookupDirector(
            atk->spawnedobj, atk->objinitstate_ref);
      else
        atk->objinitstate = atk->spawnedobj->spawn_state;
    }

    DDF_ErrorClearEntryName();
  }
}

static const specflags_t attack_specials[] =
{
    {"SMOKING TRACER", AF_TraceSmoke, 0},
    {"KILL FAILED SPAWN", AF_KillFailedSpawn, 0},
    {"REMOVE FAILED SPAWN", AF_KillFailedSpawn, 1},
    {"PRESTEP SPAWN", AF_PrestepSpawn, 0},
    {"SPAWN TELEFRAGS", AF_SpawnTelefrags, 0},
    {"NEED SIGHT", AF_NeedSight, 0},
    {"FACE TARGET", AF_FaceTarget, 0},
    {"FORCE AIM", AF_ForceAim, 0},
    {"ANGLED SPAWN", AF_AngledSpawn, 0},
    {"PLAYER ATTACK", AF_Player, 0},

    // -AJA- backwards compatibility cruft...
    {"!NOAMMO", 0, 0},
    {NULL, 0, 0}
};

static void DDF_AtkGetSpecial(const char *info, void *storage, void *priv)
{
  int flag_value;

  switch (DDF_MainCheckSpecialFlag(info, attack_specials, &flag_value,
      true, false))
  {
    case CHKF_Positive:
      buffer_atk.flags |= flag_value;
      break;
    
    case CHKF_Negative:
      buffer_atk.flags &= ~flag_value;
      break;

    case CHKF_User:
    case CHKF_Unknown:
      DDF_WarnError("DDF_AtkGetSpecials: Unknown Attack Special: %s", info);
      break;
  }
}

// -KM- 1998/11/25 Added new attack type for BFG: Spray
static const char *attack_class[NUMATKCLASS] =
{
    "PROJECTILE",
    "SPAWNER",
    "TRIPLE SPAWNER",
    "FIXED SPREADER",
    "RANDOM SPREADER",
    "SHOT",
    "TRACKER",
    "CLOSECOMBAT",
    "SHOOTTOSPOT",
    "SKULLFLY",
    "SMARTPROJECTILE",
    "SPRAY"
};

static void DDF_AtkGetType(const char *info, void *storage, void *priv)
{
  int i;

  i = 0;

  while (i != NUMATKCLASS && DDF_CompareName(info, attack_class[i]))
    i++;

  if (i == NUMATKCLASS)
  {
    DDF_WarnError("DDF_AtkGetType: No such attack type '%s'\n", info);
    buffer_atk.attackstyle = ATK_SHOT;
    return;
  }
  
  buffer_atk.attackstyle = i;
}

//
// DDF_AtkGetLabel
//
static void DDF_AtkGetLabel(const char *info, void *storage, void *priv)
{
  struct label_offset_s *lab = (struct label_offset_s *)storage;

  // check for `:' in string
  const char *div = strchr(info, DIVIDE);

  int i = div ? (div - info) : strlen(info);

  if (i <= 0)
    DDF_Error("Bad State `%s'.\n", info);

  lab->label = Z_New(const char, i + 1);
  Z_StrNCpy((char *)lab->label, info, i);

  lab->offset = div ? MAX(0, atoi(div+1) - 1) : 0;
}

//
// DDF_AttackLookup
//
attacktype_t *DDF_AttackLookup(const char *name)
{
  int i;

  if (!name || !name[0])
    DDF_Error("Missing attack name !\n");

  for (i=0; i < num_ddf_attacks; i++)
  {
    if (DDF_CompareName(ddf_attacks[i]->ddf.name, name) == 0)
      return ddf_attacks[i];
  }

  DDF_WarnError("Unknown Attack: %s\n", name);
  return NULL;
}
