//----------------------------------------------------------------------------
//  EDGE Floor/Elevator/Teleport Action Code
//----------------------------------------------------------------------------
// 
//  Copyright (c) 1999-2001  The EDGE Team.
// 
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//----------------------------------------------------------------------------
//
//  Based on the DOOM source code, released by Id Software under the
//  following copyright:
//
//    Copyright (C) 1993-1996 by id Software, Inc.
//
//----------------------------------------------------------------------------
//
// -KM-  1998/09/01 Changed for DDF.
// -ACB- 1998/09/13 Moved the teleport procedure here
//

#include "i_defs.h"

#include "z_zone.h"
#include "dm_defs.h"
#include "dm_state.h"
#include "p_local.h"
#include "r_state.h"
#include "s_sound.h"

// Linked list of moving sectors.
static elev_move_t *active_elevs = NULL;

// Linked list of moving sectors.
static sec_move_t *active_secs = NULL;

// linked list of moving sliders.
static slider_move_t *active_sliders = NULL;

static void P_RemoveActiveSector(sec_move_t * sec);
static boolean_t P_StasifySector(sector_t * sec);
static boolean_t P_ActivateInStasis(int tag);

static void MoveElevator(elev_move_t *emov);

static result_e MovePlane(sector_t * sector,
    float_t speed, float_t dest, boolean_t crush, 
    int floorOrCeiling, int direction);

static void MoveSlider(slider_move_t *smov);

// -AJA- FIXME: perhaps using a pointer to `plane_info_t' would be
//       much better that f**king about with the floorOrCeiling stuff
//       all the time.

static float_t HEIGHT(sector_t * sec, boolean_t floorOrCeiling)
{
  if (floorOrCeiling)
    return sec->ceil.h;

  return sec->floor.h;
}

static const image_t * SECPIC(sector_t * sec, boolean_t floorOrCeiling,
    const image_t *new_image)
{
  if (new_image)
  {
    if (floorOrCeiling) 
      sec->ceil.image = new_image;
    else
      sec->floor.image = new_image;
  }

  return floorOrCeiling ? sec->ceil.image : sec->floor.image;
}

#if 0  // OLD CODE
static int SECPIC(sector_t * sec, boolean_t floorOrCeiling, int newpic)
{
  if (floorOrCeiling)
  {
    if (newpic != -1)
      sec->ceil.pic = newpic;
    return sec->ceil.pic;
  }

  if (newpic != -1)
    sec->floor.pic = newpic;

  return sec->floor.pic;
}
#endif

//
// GetSecHeightReference
//
// Finds a sector height, using the reference provided; will select
// the approriate method of obtaining this value, if it cannot
// get it directly.
//
// -KM-  1998/09/01 Wrote Procedure.
// -ACB- 1998/09/06 Remarked and Reformatted.
// -ACB- 2001/02/04 Move to p_plane.c
//
static float_t GetSecHeightReference(heightref_e ref, sector_t * sec)
{
  switch (ref)
  {
    case ref_absolute:
      return 0;

    case ref_floorHeight:
      return sec->floor.h;

    case ref_ceilingHeight:
      return sec->ceil.h;

    case ref_LowestSurroundingCeiling:
      return P_FindLowestCeilingSurrounding(sec);

    case ref_HighestSurroundingCeiling:
      return P_FindHighestCeilingSurrounding(sec);

    case ref_LowestSurroundingFloor:
      return P_FindLowestFloorSurrounding(sec);

    case ref_HighestSurroundingFloor:
      return P_FindHighestFloorSurrounding(sec);

    case ref_NextHighestFloor:
      return P_FindNextHighestFloor(sec, sec->floor.h);

    case ref_NextLowestFloor:
      return P_FindNextLowestFloor(sec, sec->floor.h);

    case ref_NextHighestCeiling:
      return P_FindNextHighestCeiling(sec, sec->ceil.h);

    case ref_NextLowestCeiling:
      return P_FindNextLowestCeiling(sec, sec->ceil.h);

    case ref_LowestLoTexture:
      return P_FindRaiseToTexture(sec);

    default:
      I_Warning("GetSecHeightReference: undefined reference %d\n", ref);
  }

  return -1;
}

//
// GetElevatorHeightReference
//
// This is essentially the same as above, but used for elevator
// calculations. The reason behind this is that elevators have to
// work as dummy sectors.
//
// -ACB- 2001/02/04 Written
//
static float_t GetElevatorHeightReference(heightref_e ref, sector_t * sec)
{
  return -1;
}

//
// FLOORS
//

//
// MoveSector
//
// MOVE A FLOOR TO IT'S DESTINATION (UP OR DOWN)
//
static void MoveSector(sec_move_t * sec)
{
  result_e res;

  switch (sec->direction)
  {
      // In Stasis
    case -2:
      sec->sfxstarted = false;
      break;

      // DOWN
    case -1:
      res = MovePlane(sec->sector,
          sec->speed,
          sec->startheight < sec->destheight ?
          sec->startheight : sec->destheight,
          sec->floorOrCeiling && sec->crush,
          sec->floorOrCeiling, sec->direction);

      if (!sec->sfxstarted)
      {
        S_StartSound((mobj_t *) & sec->sector->soundorg, sec->type->sfxdown);
        sec->sfxstarted = true;
      }

      if (res == RES_PastDest)
      {
        S_StartSound((mobj_t *) & sec->sector->soundorg, sec->type->sfxstop);
        sec->speed = sec->type->speed_up;

        if (sec->newspecial != -1)
        {
          sec->sector->p.special = (sec->newspecial <= 0) ? NULL :
            DDF_SectorLookupNum(sec->newspecial);
        }

        SECPIC(sec->sector, sec->floorOrCeiling, sec->new_image);

        switch (sec->type->type)
        {
          case mov_Plat:
          case mov_Continuous:
            sec->direction = 0;
            sec->waited = sec->type->wait;
            sec->speed = sec->type->speed_up;
            break;

          case mov_MoveWaitReturn:
            if (HEIGHT(sec->sector, sec->floorOrCeiling) == sec->startheight && sec->completed)
            {
              P_RemoveActiveSector(sec);
            }
            else if (HEIGHT(sec->sector, sec->floorOrCeiling) == sec->destheight)
            {
              sec->direction = 0;
              sec->waited = sec->type->wait;
              sec->speed = sec->type->speed_up;
              sec->completed = true;
            }
            break;

          default:
          case mov_Stairs:
          case mov_Once:
            P_RemoveActiveSector(sec);
            break;
        }
      }
      else if (res == RES_Crushed)
      {
        if (sec->crush)
        {
          sec->speed = sec->type->speed_down / 8;
        }
        else if (sec->type->type == mov_MoveWaitReturn)  // Go back up
        {
          sec->direction = 1;
          sec->sfxstarted = false;
          sec->waited = 0;
          sec->speed = sec->type->speed_up;
        }
      }

      break;

    // WAITING
    case 0:
      if (--sec->waited <= 0)
      {
        int dir;
        float_t dest;

        if (HEIGHT(sec->sector, sec->floorOrCeiling) == sec->destheight)
          dest = sec->startheight;
        else
          dest = sec->destheight;

        if (HEIGHT(sec->sector, sec->floorOrCeiling) > dest)
        {
          dir = -1;
          sec->speed = sec->type->speed_down;
        }
        else
        {
          dir = 1;
          sec->speed = sec->type->speed_up;
        }

        sec->direction = dir;  // time to go back

        if (dir)
        {
          S_StartSound((mobj_t *) &sec->sector->soundorg,
                                                sec->type->sfxstart);
        }

        sec->sfxstarted = false;
      }
      break;

    // UP
    case 1:
      res = MovePlane(sec->sector,
          sec->speed,
          sec->destheight > sec->startheight ?
          sec->destheight : sec->startheight,
          sec->crush && !sec->floorOrCeiling,
          sec->floorOrCeiling, sec->direction);

      if (!sec->sfxstarted)
      {
        S_StartSound((mobj_t *) & sec->sector->soundorg, sec->type->sfxup);
        sec->sfxstarted = true;
      }

      if (res == RES_PastDest)
      {
        S_StartSound((mobj_t *) & sec->sector->soundorg, sec->type->sfxstop);

        if (sec->newspecial != -1)
        {
          sec->sector->p.special = (sec->newspecial <= 0) ? NULL :
            DDF_SectorLookupNum(sec->newspecial);
        }

        SECPIC(sec->sector, sec->floorOrCeiling, sec->new_image);

        switch (sec->type->type)
        {
          case mov_Plat:
          case mov_Continuous:
            sec->direction = 0;
            sec->waited = sec->type->wait;
            sec->speed = sec->type->speed_down;
            break;

          case mov_MoveWaitReturn:
            if (sec->completed &&
                HEIGHT(sec->sector, sec->floorOrCeiling) == sec->startheight)
            {
              P_RemoveActiveSector(sec);
            }
            else if (HEIGHT(sec->sector, sec->floorOrCeiling) == sec->destheight)
            {
              sec->direction = 0;
              sec->speed = sec->type->speed_down;
              sec->waited = sec->type->wait;
              sec->completed = true;
            }
            break;

          default:
          case mov_Once:
          case mov_Stairs:
            P_RemoveActiveSector(sec);
            break;
        }

      }
      else if (res == RES_Crushed)
      {
        if (sec->crush)
        {
          sec->speed = sec->type->speed_up / 8;
        }
        else if (sec->type->type == mov_MoveWaitReturn)  // Go back down
        {
          sec->direction = -1;
          sec->sfxstarted = false;
          sec->waited = 0;
          sec->speed = sec->type->speed_down;
        }
      }
      break;
    default:
      I_Error("MoveSector: Unknown direction %d", sec->direction);
  }

}

//
// P_RunActiveSectors
//
// Executes one tic's sec_move_t thinking.
// Active sectors can destroy themselves, but not each other.
// We do not have to bother about a removal queue, but we can not rely on
// sec still being in memory after MoveSector.
//
// -AJA- 2000/08/06 Now handles horizontal sliding doors too.
// -ACB- 2001/01/14 Now handles elevators too.
//
void P_RunActiveSectors(void)
{
  elev_move_t *elev, *elev_next;
  sec_move_t *sec, *sec_next;
  slider_move_t *slid, *slid_next;

  for (elev = active_elevs; elev; elev = elev_next)
  {
    elev_next = elev->next;
    MoveElevator(elev);
  }

  for (sec = active_secs; sec; sec = sec_next)
  {
    sec_next = sec->next;
    MoveSector(sec);
  }

  for (slid = active_sliders; slid; slid = slid_next)
  {
    slid_next = slid->next;
    MoveSlider(slid);
  }
}

//
// P_GSS
//
static sector_t *P_GSS(sector_t * sec, float_t dest, boolean_t forc)
{
  int i;
  int secnum = sec - sectors;
  sector_t *sector;

  for (i = sec->linecount-1; i; i--)
  {
    if (P_TwoSided(secnum, i))
    {
      if (P_GetSide(secnum, i, 0)->sector - sectors == secnum)
      {
        sector = P_GetSector(secnum, i, 1);

        if (SECPIC(sector, forc, NULL) != SECPIC(sec, forc, NULL)
            && HEIGHT(sector, forc) == dest)
        {
          return sector;
        }

      }
      else
      {
        sector = P_GetSector(secnum, i, 0);
        
        if (SECPIC(sector, forc, NULL) != SECPIC(sec, forc, NULL)
            && HEIGHT(sector, forc) == dest)
        {
          return sector;
        }
      }
    }
  }

  for (i = sec->linecount; i--;)
  {
    if (P_TwoSided(secnum, i))
    {
      if (P_GetSide(secnum, i, 0)->sector - sectors == secnum)
      {
        sector = P_GetSector(secnum, i, 1);
      }
      else
      {
        sector = P_GetSector(secnum, i, 0);
      }
      if (sector->validcount != validcount)
      {
        sector->validcount = validcount;
        sector = P_GSS(sector, dest, forc);
        if (sector)
          return sector;
      }
    }
  }

  return NULL;
}

static sector_t *P_GetSectorSurrounding(sector_t * sec, float_t dest, boolean_t forc)
{
  validcount++;
  sec->validcount = validcount;
  return P_GSS(sec, dest, forc);
}

//
// P_SetupSectorAction
//
// Setup the Floor Action, depending on the linedeftype trigger and the
// sector info.
//
static sec_move_t *P_SetupSectorAction(sector_t * sector, const moving_plane_t * type, sector_t * model)
{
  sec_move_t *sec;
  float_t start, dest;

  // new door thinker
  sec = Z_New(sec_move_t, 1);

  if (type->floorOrCeiling)
    sector->ceil.movedata = (void*)sec;
  else
    sector->floor.movedata = (void*)sec;

  sec->whatiam = MDT_PLANE;
  sec->sector = sector;
  sec->crush = type->crush;
  sec->sfxstarted = sec->completed = false;
  start = HEIGHT(sector, type->floorOrCeiling);

  dest = GetSecHeightReference(type->destref, sector);
  dest += type->dest;

  if (type->type == mov_Plat)
    start = GetSecHeightReference(ref_HighestSurroundingFloor, sector);

  //--------------------------------------------------------------------------
  // Floor Speed Notes:
  //
  //   Floor speed setup; -M_PI is the default speed, This is used to simulate
  //   the use of an instant movement. i.e. a floor that raises or falls to
  //   its destination height in one tic: This is also implemented for WAD's
  //   that use odd linedef requests to achieve the instant effect.
  //
  //   If someone would want a speed close to -3.1, the probability that
  //   he would use -M_PI accidentally is really low, so we do not care about
  //   it. I can't think of any situation where there is a point in setting
  //   the speed to exactly pi.
  //
  //   Therefore a speed of -M_PI, is translated to the distance between the
  //   start and destination: instant movement; otherwise the speed is taken
  //   from the linedef type.
  //
  //--------------------------------------------------------------------------
  if (type->prewait)
  {
    sec->direction = 0;
    sec->waited = type->prewait;
  }
  else if (dest > start)
  {
    sec->direction = 1;

    // -ACB- 1998/09/09 See floor speed notes...
    if (type->speed_up >= 0)
      sec->speed = type->speed_up;
    else
      sec->speed = dest - start;
  }
  else if (start > dest)
  {
    sec->direction = -1;

    // -ACB- 1998/09/09 See floor speed notes...
    if (type->speed_down >= 0)
      sec->speed = type->speed_down;
    else
      sec->speed = start - dest;
  }
  else
  {
    if (type->floorOrCeiling)
      sector->ceil.movedata = NULL;
    else
      sector->floor.movedata = NULL;

    Z_Free(sec);
    return NULL;
  }

  sec->destheight = dest;
  sec->startheight = start;
  sec->tag = sector->tag;
  sec->type = type;
  sec->new_image = SECPIC(sector, type->floorOrCeiling, NULL);
  sec->newspecial = -1;
  sec->floorOrCeiling = type->floorOrCeiling;

  // -ACB- 10/01/2001 Trigger starting sfx
  S_StopLoopingSound((mobj_t *) & sector->soundorg);
  if (type->sfxstart)
    S_StartSound((mobj_t *) & sector->soundorg, type->sfxstart);

  // change to surrounding
  if (type->tex[0] == '-')
  {
    model = P_GetSectorSurrounding(sector, sec->destheight, type->floorOrCeiling);
    if (model)
    {
      sec->new_image = SECPIC(model, type->floorOrCeiling, NULL);

      sec->newspecial = model->p.special ?
          model->p.special->ddf.number : 0;
    }
    if (sec->direction == (type->floorOrCeiling ? -1 : 1))
    {
      SECPIC(sector, type->floorOrCeiling, sec->new_image);
      if (sec->newspecial != -1)
      {
        sector->p.special = (sec->newspecial <= 0) ? NULL :
          DDF_SectorLookupNum(sec->newspecial);
      }
    }
  }
  else if (type->tex[0] == '+')
  {
    if (model)
    {
      if (SECPIC(model,  type->floorOrCeiling, NULL) == 
          SECPIC(sector, type->floorOrCeiling, NULL))
      {
        model = P_GetSectorSurrounding(model, sec->destheight,
          type->floorOrCeiling);
      }
    }

    if (model)
    {
      sec->new_image = SECPIC(model, type->floorOrCeiling, NULL);
      sec->newspecial = model->p.special ?
          model->p.special->ddf.number : 0;

      if (sec->direction == (type->floorOrCeiling ? -1 : 1))
      {
        SECPIC(sector, type->floorOrCeiling, sec->new_image);

        if (sec->newspecial != -1)
        {
          sector->p.special = (sec->newspecial <= 0) ? NULL :
            DDF_SectorLookupNum(sec->newspecial);
        }
      }
    }
  }
  else if (type->tex[0])
  {
    sec->new_image = W_ImageFromFlat(type->tex);
  }

  P_AddActiveSector(sec);
  return sec;
}

//
// EV_Teleport
//
// Teleportation is an effect which is simulated by searching for the first
// special[MOBJ_TELEPOS] in a sector with the same tag as the activation line,
// moving an object from one sector to another upon the MOBJ_TELEPOS found, and
// possibly spawning an effect object (i.e teleport flash) at either the entry &
// exit points or both.
//
// -KM- 1998/09/01 Added stuff for lines.ddf (mostly sounds)
//
// -ACB- 1998/09/11 Reformatted and cleaned up.
//
// -ACB- 1998/09/12 Teleport delay setting from linedef.
//
// -ACB- 1998/09/13 used effect objects: the objects themselves make any sound and
//                  the in effect object can be different to the out object.
//
// -ACB- 1998/09/13 Removed the missile checks: no need since this would have been
//                  Checked at the linedef stage.
//
// -KM- 1998/11/25 Changed Erik's code a bit, Teleport flash still appears.
//  if def faded_teleportation == 1, doesn't if faded_teleportation == 2
//
// -ES- 1998/11/28 Changed Kester's code a bit :-) Teleport method can now be
//  toggled in the menu. (That is the way it should be. -KM)
//
// -KM- 1999/01/31 Search only the target sector, not the entire map.
//
// -AJA- 1999/07/12: Support for TELEPORT_SPECIAL in lines.ddf.
// -AJA- 1999/07/30: Updated for extra floor support.
// -AJA- 1999/10/21: Allow line to be NULL, and added `tag' param.
//
boolean_t EV_Teleport
(
  line_t* line,
  int tag, 
  int side,
  mobj_t* thing,
  int delay,
  int special,
  const mobjinfo_t* ineffectobj,
  const mobjinfo_t * outeffectobj
)
{
  int i;
  angle_t an;
  mobj_t *currmobj;
  float_t oldx;
  float_t oldy;
  float_t oldz;
  float_t centre_x, centre_y;
  float_t new_x, new_y, new_z;
  mobj_t *fog;

  if (!thing)
    return false;

  for (i = 0; i < numsubsectors; i++)
  {
    if (subsectors[i].sector->tag != tag)
      continue;

    currmobj = subsectors[i].thinglist;

    while (currmobj)
    {
      // not a teleportman
      if (currmobj->info != outeffectobj)
      {
        currmobj = currmobj->snext;
        continue;
      }

      oldx = thing->x;
      oldy = thing->y;
      oldz = thing->z;

      new_x = currmobj->x;
      new_y = currmobj->y;
      new_z = currmobj->z;

      if (line && (special & TELSP_SameOffset))
      {
        centre_x = (line->v1->x + line->v2->x) / 2;
        centre_y = (line->v1->y + line->v2->y) / 2;

        new_x += thing->x - centre_x;
        new_y += thing->y - centre_y;
      }

      if (special & TELSP_SameHeight)
        new_z += (thing->z - thing->floorz);
      else if (thing->flags & MF_MISSILE)
        new_z += thing->origheight;

      if (!P_TeleportMove(thing, new_x, new_y, new_z))
        return false;

      if (thing->player)
        thing->player->viewz = thing->z + thing->player->viewheight;

      // spawn teleport fog at source and destination
      if (ineffectobj)
      {
        fog = P_MobjCreateObject(oldx, oldy, oldz, ineffectobj);

        if (fog->info->chase_state)
          P_SetMobjState(fog, fog->info->chase_state);
      }

      an = currmobj->angle;

      //
      // -ACB- 1998/09/06 Switched 40 to 20. This by my records is
      //                  the original setting.
      //
      // -ES- 1998/10/29 When fading, we don't want to see the fog.
      //
      fog = P_MobjCreateObject(currmobj->x + 20 * M_Cos(an),
          currmobj->y + 20 * M_Sin(an),
          currmobj->z, outeffectobj);

      if (fog->info->chase_state)
        P_SetMobjState(fog, fog->info->chase_state);

      if (thing->player && !telept_flash)
        fog->vis_target = fog->visibility = INVISIBLE;

      // don't move for a bit
      if (thing->player && !(special & TELSP_SameSpeed))
      {
        thing->reactiontime = delay;
        // -ES- 1998/10/29 Start the fading
        if (telept_effect && thing->player == displayplayer)
          R_StartFading(0, (delay * 5) / 2);
        thing->momx = thing->momy = thing->momz = 0;
      }

      if (special & TELSP_Rotate)
        thing->angle += currmobj->angle;
      else if (!(special & TELSP_SameDir))
      {
        thing->angle = currmobj->angle;
        thing->vertangle = currmobj->vertangle;
      }

      if (thing->flags & MF_MISSILE)
      {
        thing->momx = thing->speed * M_Cos(thing->angle);
        thing->momy = thing->speed * M_Sin(thing->angle);
      }

      return true;

    }  // while (currmobj)

  }  // for (subsector) loop

  return false;
}

//
// EV_BuildOneStair
//
// BUILD A STAIRCASE!
//
// -AJA- 1999/07/04: Fixed the problem on MAP20. The next stair's
// dest height should be relative to the previous stair's dest height
// (and not just the current height).
//
// -AJA- 1999/07/29: Split into two functions. The old code could do bad
// things (e.g. skip a whole staircase) when 2 or more stair sectors
// were tagged.
//
static boolean_t EV_BuildOneStair(sector_t * sec, const moving_plane_t * type)
{
  int i;
  float_t next_height;
  boolean_t more;
  boolean_t rtn;

  sec_move_t *stairs;
  sector_t *tsec;
  float_t stairsize = type->dest;

  const image_t *image = sec->floor.image;

  // new floor thinker

  stairs = P_SetupSectorAction(sec, type, sec);
  rtn = stairs ? true : false;
  next_height = stairs->destheight + stairsize;

  do
  {
    more = false;

    // Find next sector to raise
    //
    // 1. Find 2-sided line with same sector side[0]
    // 2. Other side is the next sector to raise
    //
    for (i = 0; i < sec->linecount; i++)
    {
      if (!(sec->lines[i]->flags & ML_TwoSided))
        continue;

      if (sec != sec->lines[i]->frontsector)
        continue;

      tsec = sec->lines[i]->backsector;

      if (tsec->floor.image != image)
        continue;

      if (type->floorOrCeiling && tsec->ceil.movedata)
        continue;

      if (!type->floorOrCeiling && tsec->floor.movedata)
        continue;

      stairs = P_SetupSectorAction(tsec, type, tsec);

      if (stairs)
      {
        stairs->destheight = next_height;
        next_height += stairsize;
        sec = tsec;
        more = true;
      }

      break;
    }
  }
  while (more);

  return rtn;
}

//
// EV_BuildStairs
//
static boolean_t EV_BuildStairs(sector_t * sec, const moving_plane_t * type)
{
  boolean_t rtn = false;

  while (sec->tag_prev)
    sec = sec->tag_prev;

  for (; sec; sec = sec->tag_next)
  {
    // Already moving?  If so, keep going...
    if (sec->ceil.movedata && type->floorOrCeiling)
      continue;

    // Already moving?  If so, keep going...
    if (sec->floor.movedata && !type->floorOrCeiling)
      continue;

    if (EV_BuildOneStair(sec, type))
      rtn = true;
  }

  return rtn;
}

//
// EV_DoSector
//
// Do Platforms/Floors/Stairs/Ceilings/Doors
//
boolean_t EV_DoSector(sector_t * sec, const moving_plane_t * type, sector_t * model)
{
  // Activate all <type> plats that are in_stasis
  switch (type->type)
  {
    case mov_Plat:
    case mov_Continuous:
      if (P_ActivateInStasis(sec->tag))
        return true;
      break;

    case mov_Stairs:
      return EV_BuildStairs(sec, type);

    case mov_Stop:
      return P_StasifySector(sec);

    default:
      break;
  }

  if (type->floorOrCeiling)
  {
    if (sec->ceil.movedata)
      return false;
  }
  else
  {
    if (sec->floor.movedata)
      return false;
  }

  // Do Floor action
  return P_SetupSectorAction(sec, type, model) ? true : false;
}

//
// EV_ManualPlane
//
boolean_t EV_ManualPlane(line_t * line, mobj_t * thing, const moving_plane_t * type)
{
  sector_t *sec;
  sec_move_t *msec;
  int side;
  int dir = 1;
  int olddir = 1;

  side = 0;  // only front sides can be used

  // if the sector has an active thinker, use it
  sec = side ? line->frontsector : line->backsector;
  if (!sec)
    return false;

  if (type->floorOrCeiling)
    msec = sec->ceil.movedata;
  else
    msec = sec->floor.movedata;

  if (msec && thing)
  {
    switch (type->type)
    {
      case mov_MoveWaitReturn:
        olddir = msec->direction;

        // Only players close doors
        if ((msec->direction != -1) && thing->player)
          dir = msec->direction = -1;
        else
          dir = msec->direction = 1;
        break;
        
      default:
        break;
    }

    if (dir != olddir)
    {
      S_StartSound((mobj_t *) & sec->soundorg, type->sfxstart);
      msec->sfxstarted = !(thing->player);
      return true;
    }

    return false;
  }

  return EV_DoSector(sec, type, sec);
}

//
// P_ActivateInStasis
//
static boolean_t P_ActivateInStasis(int tag)
{
  sec_move_t *s;
  boolean_t rtn = false;

  for (s = active_secs; s; s = s->next)
  {
    if(s->direction == -2 && s->tag == tag)
    {
      s->direction = s->olddirection;
      rtn = true;
    }
  }

  return rtn;
}

//
// P_StasifySector
//
static boolean_t P_StasifySector(sector_t * sec)
{
  sec_move_t *s;
  boolean_t rtn = false;

  for (s = active_secs; s; s = s->next)
  {
    if (s->direction != -2 && s->tag == sec->tag)
    {
      s->olddirection = s->direction;
      s->direction = -2;
      rtn = true;
    }
  }

  return rtn;
}

//
// P_AddActiveSector
//
void P_AddActiveSector(sec_move_t * sec)
{
  // -ES- Insert into the tail.
  sec_move_t *tmp;

  sec->next = NULL;
  if (!active_secs)
  {
    active_secs = sec;
    sec->prev = NULL;
    return;
  }
  
  for(tmp = active_secs; tmp->next; tmp = tmp->next)
    ;

  tmp->next = sec;
  sec->prev = tmp;
}

//
// P_RemoveAllActiveSectors
//
void P_RemoveAllActiveSectors(void)
{
  sec_move_t *sec, *next;

  for (sec = active_secs; sec; sec = next)
  {
    next = sec->next;
    Z_Free(sec);
  }
  
  active_secs = NULL;
}

//
// P_RemoveActiveSector
//
static void P_RemoveActiveSector(sec_move_t * sec)
{
  if (sec->floorOrCeiling)
    sec->sector->ceil.movedata = NULL;
  else
    sec->sector->floor.movedata = NULL;

  if (sec->prev)
    sec->prev->next = sec->next;
  else
    active_secs = sec->next;

  if (sec->next)
    sec->next->prev = sec->prev;
    
  Z_Free(sec);
}


//
// MovePlane
//
// Move a plane (floor or ceiling) and check for crushing
//
static result_e MovePlane(sector_t * sector, 
    float_t speed, float_t dest, boolean_t crush, 
    int floorOrCeiling, int direction)
{
  boolean_t flag;
  float_t lastpos;

  // -AJA- Floors that move down, and ceilings that move up, do not
  // normally cause any crushing (the gap is widening, not narrowing).
  // Therefore we shouldn't stop a floor from lowering or a ceiling
  // from rising if something "gets in the way". 
  //
  // But when the sector is controlling an extra floor, then the
  // actual gap may indeed be narrowing.  Hence this flag.

  boolean_t extra = sector->controller;

  switch (floorOrCeiling)
  {
    case 0:
      // FLOOR
      switch (direction)
      {
        case -1:
          // DOWN
          if (sector->floor.h - speed < dest)
          {
            lastpos = sector->floor.h;
            sector->floor.h = dest;
            flag = P_ChangeSector(sector, crush);

            if (extra && flag && !crush)
            {
              sector->floor.h = lastpos;
              P_ChangeSector(sector, crush);
            }
            return RES_PastDest;
          }
          else
          {
            lastpos = sector->floor.h;
            sector->floor.h -= speed;
            flag = P_ChangeSector(sector, crush);

            if (extra && flag && !crush)
            {
              sector->floor.h = lastpos;
              P_ChangeSector(sector, crush);
            }
            return flag ? RES_Crushed : RES_Ok;
          }

        case 1:
          // UP
          if (sector->floor.h + speed > dest)
          {
            lastpos = sector->floor.h;
            sector->floor.h = dest;
            flag = P_ChangeSector(sector, crush);

            if (flag && !crush)
            {
              sector->floor.h = lastpos;
              P_ChangeSector(sector, crush);
            }
            return RES_PastDest;
          }
          else
          {
            lastpos = sector->floor.h;
            sector->floor.h += speed;
            flag = P_ChangeSector(sector, crush);

            if (flag && !crush)
            {
              sector->floor.h = lastpos;
              P_ChangeSector(sector, crush);
            }
            return flag ? RES_Crushed : RES_Ok;
          }
      }
      break;

    case 1:
      // CEILING
      switch (direction)
      {
        case -1:
          // DOWN
          if (sector->ceil.h - speed < dest)
          {
            lastpos = sector->ceil.h;
            sector->ceil.h = dest;
            flag = P_ChangeSector(sector, crush);

            if (flag && !crush)
            {
              sector->ceil.h = lastpos;
              P_ChangeSector(sector, crush);
            }
            return RES_PastDest;
          }
          else
          {
            lastpos = sector->ceil.h;
            sector->ceil.h -= speed;
            flag = P_ChangeSector(sector, crush);

            if (flag && !crush)
            {
              sector->ceil.h = lastpos;
              P_ChangeSector(sector, crush);
            }
            return flag ? RES_Crushed : RES_Ok;
          }

        case 1:
          // UP
          if (sector->ceil.h + speed > dest)
          {
            lastpos = sector->ceil.h;
            sector->ceil.h = dest;
            flag = P_ChangeSector(sector, crush);

            if (extra && flag && !crush)
            {
              sector->ceil.h = lastpos;
              P_ChangeSector(sector, crush);
            }
            return RES_PastDest;
          }
          else
          {
            lastpos = sector->ceil.h;
            sector->ceil.h += speed;
            flag = P_ChangeSector(sector, crush);

            if (extra && flag && !crush)
            {
              sector->ceil.h = lastpos;
              P_ChangeSector(sector, crush);
            }
            return flag ? RES_Crushed : RES_Ok;
          }
      }
      break;
  }
  return RES_Ok;
}

// -AJA- 1999/12/07: cleaned up this donut stuff

static linedeftype_t donut[2];
static int donut_setup = 0;

//
// EV_DoDonut
//
// Special Stuff that can not be categorized
// Mmmmmmm....  Donuts....
//
boolean_t EV_DoDonut(sector_t * s1, sfx_t *sfx[4])
{
  sector_t *s2;
  sector_t *s3;
  boolean_t result = false;
  int i;
  sec_move_t *sec;

  if (! donut_setup)
  {
    donut[0] = template_line;
    donut[0].count = 1;
    donut[0].specialtype = 0;
    donut[0].f = donut_floor;
    strcpy(donut[0].f.tex, "-");

    donut[1] = template_line;
    donut[1].count = 1;
    donut[1].f = donut_floor;
    donut[1].f.dest = INT_MIN;

    donut_setup++;
  }
  
  // ALREADY MOVING?  IF SO, KEEP GOING...
  if (s1->floor.movedata)
    return false;

  s2 = P_GetNextSector(s1->lines[0], s1);

  for (i = 0; i < s2->linecount; i++)
  {
    if ((!s2->lines[i]->flags & ML_TwoSided) || (s2->lines[i]->backsector == s1))
      continue;

    s3 = s2->lines[i]->backsector;

    result = true;

    // Spawn rising slime
    donut[0].f.sfxup = sfx[0];
    donut[0].f.sfxstop = sfx[1];
    
    sec = P_SetupSectorAction(s2, &donut[0].f, s3);

    if (sec)
    {
      sec->destheight = s3->floor.h;
      s2->floor.image = sec->new_image = s3->floor.image;
      s2->p.special = s3->p.special;
    }

    // Spawn lowering donut-hole
    donut[1].f.sfxup = sfx[2];
    donut[1].f.sfxstop = sfx[3];

    sec = P_SetupSectorAction(s1, &donut[1].f, s1);

    if (sec)
      sec->destheight = s3->floor.h;
    break;
  }

  return result;
}


//
// P_AddActiveSlider
//
// HORIZONTAL SLIDER STUFF
//
void P_AddActiveSlider(slider_move_t *smov)
{
  // -ES- Insert into the tail.
  slider_move_t *tmp;

  smov->next = NULL;

  if (! active_sliders)
  {
    active_sliders = smov;
    smov->prev = NULL;
    return;
  }
  
  for (tmp = active_sliders; tmp->next; tmp = tmp->next)
  { /* nothing here */ }

  tmp->next = smov;
  smov->prev = tmp;
}

//
// P_RemoveAllActiveSliders
//
// This routine meant for level-shutdown code.
//
void P_RemoveAllActiveSliders(void)
{
  slider_move_t *smov, *next;

  for (smov = active_sliders; smov; smov = next)
  {
    next = smov->next;
    Z_Free(smov);
  }
 
  active_sliders = NULL;
}

//
// P_RemoveActiveSlider
//
static void P_RemoveActiveSlider(slider_move_t *smov)
{
  smov->line->slider_special = NULL;

  if (smov->prev)
    smov->prev->next = smov->next;
  else
    active_sliders = smov->next;

  if (smov->next)
    smov->next->prev = smov->prev;
    
  Z_Free(smov);
}

//
// SliderCanClose
//
static INLINE boolean_t SliderCanClose(line_t *line)
{
  return ! P_ThingsOnLine(line);
}

//
// MoveSlider
//
static void MoveSlider(slider_move_t *smov)
{
  sector_t *sec = smov->line->frontsector;

  switch (smov->direction)
  {
    // WAITING
    case 0:
      if (--smov->waited <= 0)
      {
        if (SliderCanClose(smov->line))
        {
          S_StartSound((mobj_t *) & sec->soundorg, smov->info->sfx_start);
          smov->sfxstarted = false;
          smov->direction = -1;
        }
        else
        {
          // try again soon
          smov->waited = TICRATE / 3;
        }
      }
      break;

    // OPENING
    case 1:
      if (! smov->sfxstarted)
      {
        S_StartSound((mobj_t *) & sec->soundorg, smov->info->sfx_open);
        smov->sfxstarted = true;
      }

      smov->opening += smov->info->speed;

      // mark line as non-blocking (at some point)
      P_ComputeGaps(smov->line);

      if (smov->opening >= smov->target)
      {
        S_StartSound((mobj_t *) & sec->soundorg, smov->info->sfx_stop);
        smov->opening = smov->target;
        smov->direction = 0;
        smov->waited = smov->info->wait;

        if (smov->final_open)
        {
          // clear line special
          smov->line->special = NULL;

          // clear the side textures
          smov->line->side[0]->middle.image = NULL;
          smov->line->side[1]->middle.image = NULL;

          P_RemoveActiveSlider(smov);
          return;
        }
      }
      break;

    // CLOSING
    case -1:
      if (! smov->sfxstarted)
      {
        S_StartSound((mobj_t *) & sec->soundorg, smov->info->sfx_close);
        smov->sfxstarted = true;
      }

      smov->opening -= smov->info->speed;

      // mark line as blocking (at some point)
      P_ComputeGaps(smov->line);

      if (smov->opening <= 0.0)
      {
        S_StartSound((mobj_t *) & sec->soundorg, smov->info->sfx_stop);
        P_RemoveActiveSlider(smov);
        return;
      }
      break;

    default:
      I_Error("MoveSlider: Unknown direction %d", smov->direction);
  }
}

//
// EV_Slider
//
// Handle thin horizontal sliding doors.
//
boolean_t EV_Slider(line_t * line, mobj_t * thing, const sliding_door_t * s)
{
  sector_t *sec = line->frontsector;
  slider_move_t *smov;

  if (! thing || ! sec || ! line->side[0] || ! line->side[1])
    return false;

  // if the line has an active thinker, use it
  if (line->slider_special)
  {
    smov = line->slider_special;

    // only players close doors
    if (smov->direction == 0 && thing->player)
    {
      smov->waited = 0;
      return true;
    }
    return false;
  }

  // new sliding door thinker
  smov = Z_New(slider_move_t, 1);

  smov->whatiam = MDT_SLIDER;
  smov->info = &line->special->s;
  smov->line = line;
  smov->opening = 0.0;
  smov->target  = R_PointToDist(0, 0, line->dx, line->dy) *
      PERCENT_2_FLOAT(smov->info->distance);

  smov->direction = 1;
  smov->sfxstarted = ! thing->player;
  smov->final_open = (line->count == 1);

  line->slider_special = smov;

  P_AddActiveSlider(smov);

  S_StartSound((mobj_t *) & sec->soundorg, s->sfx_start);
  return true;
}

//
// MoveElevator
//
static void MoveElevator(elev_move_t *elevator)
{
  result_e res;

  switch (elevator->direction)
  {
    default:
      break;
  }

  return;
}

//
// EV_DoElevator
//
// Do Elevators
//
boolean_t EV_DoElevator(sector_t * sec, const elevator_sector_t * type, sector_t * model)
{
/*
  // Activate all <type> plats that are in_stasis
  switch (type->type)
  {
    case mov_Plat:
    case mov_Continuous:
      if (P_ActivateInStasis(sec->tag))
        return true;

      break;

    case mov_Stop:
      return P_StasifySector(sec);

    default:
      break;
  }
*/

  if (!sec->controller)
    return false;

  if (sec->ceil.movedata || sec->floor.movedata)
    return false;

  // Do Elevator action
  return P_SetupElevatorAction(sec, type, model) ? true : false;
}

//
// EV_ManualElevator
//
boolean_t EV_ManualElevator(line_t * line, mobj_t * thing,
                                        const elevator_sector_t * type)
{
  return false;
}

//
// P_SetupElevatorAction
//
static elev_move_t *P_SetupElevatorAction(sector_t * sector,
                        const elevator_sector_t * type, sector_t * model)
{
  elev_move_t *elev;
  float_t start, dest;

  // new door thinker
  elev = Z_New(elev_move_t, 1);

  sector->ceil.movedata = (void*)elev;
  sector->floor.movedata = (void*)elev;

  elev->whatiam = MDT_ELEVATOR;
  elev->sector = sector;
  elev->sfxstarted = elev->completed = false;

//???  start = sector->ceil.h;

  dest = GetElevatorHeightReference(type->destref, sector);
  dest += type->dest;

/*

  if (type->type == mov_Plat)
    start = GetSecHeightReference(ref_HighestSurroundingFloor, sector);

  if (type->prewait)
  {
    elev->direction = 0;
    elev->waited = type->prewait;
  }
  else if (dest > start)
  {
    elev->direction = 1;

    // -ACB- 1998/09/09 See floor speed notes...
    if (type->speed_up >= 0)
      sec->speed = type->speed_up;
    else
      sec->speed = dest - start;
  }
  else if (start > dest)
  {
    sec->direction = -1;

    // -ACB- 1998/09/09 See floor speed notes...
    if (type->speed_down >= 0)
      sec->speed = type->speed_down;
    else
      sec->speed = start - dest;
  }
  else
  {
    sector->ceil.movedata = NULL;
    sector->floor.movedata = NULL;

    Z_Free(elev);
    return NULL;
  }

  sec->destheight = dest;
  sec->startheight = start;
  sec->tag = sector->tag;
  sec->type = type;
  sec->new_image = SECPIC(sector, type->floorOrCeiling, NULL);
  sec->newspecial = -1;
  sec->floorOrCeiling = type->floorOrCeiling;

  // -ACB- 10/01/2001 Trigger starting sfx
  S_StopLoopingSound((mobj_t *) & sector->soundorg);
  if (type->sfxstart)
    S_StartSound((mobj_t *) & sector->soundorg, type->sfxstart);

  // change to surrounding
  if (type->tex[0] == '-')
  {
    model = P_GetSectorSurrounding(sector, sec->destheight, type->floorOrCeiling);
    if (model)
    {
      sec->new_image = SECPIC(model, type->floorOrCeiling, NULL);

      sec->newspecial = model->p.special ?
          model->p.special->ddf.number : 0;
    }
    if (sec->direction == (type->floorOrCeiling ? -1 : 1))
    {
      SECPIC(sector, type->floorOrCeiling, sec->new_image);
      if (sec->newspecial != -1)
      {
        sector->p.special = (sec->newspecial <= 0) ? NULL :
          DDF_SectorLookupNum(sec->newspecial);
      }
    }
  }
  else if (type->tex[0] == '+')
  {
    if (model)
    {
      if (SECPIC(model,  type->floorOrCeiling, NULL) == 
          SECPIC(sector, type->floorOrCeiling, NULL))
      {
        model = P_GetSectorSurrounding(model, sec->destheight,
          type->floorOrCeiling);
      }
    }

    if (model)
    {
      sec->new_image = SECPIC(model, type->floorOrCeiling, NULL);
      sec->newspecial = model->p.special ?
          model->p.special->ddf.number : 0;

      if (sec->direction == (type->floorOrCeiling ? -1 : 1))
      {
        SECPIC(sector, type->floorOrCeiling, sec->new_image);

        if (sec->newspecial != -1)
        {
          sector->p.special = (sec->newspecial <= 0) ? NULL :
            DDF_SectorLookupNum(sec->newspecial);
        }
      }
    }
  }
  else if (type->tex[0])
  {
    sec->new_image = W_ImageFromFlat(type->tex);
  }

  P_AddActiveSector(sec);
  return sec;
*/
  return NULL;
}
