//----------------------------------------------------------------------------
//  EDGE Rendering Data Handling Code
//----------------------------------------------------------------------------
// 
//  Copyright (c) 1999-2000  The EDGE Team.
// 
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//----------------------------------------------------------------------------
//
//  Based on the DOOM source code, released by Id Software under the
//  following copyright:
//
//    Copyright (C) 1993-1996 by id Software, Inc.
//
//----------------------------------------------------------------------------
//
// -ACB- 1998/09/09 Reformatted File Layout.
// -KM- 1998/09/27 Colourmaps can be dynamically changed.
// -ES- 2000/02/12 Moved most of this module to w_texture.c.

#include "i_defs.h"
#include "r_data.h"

#include "e_search.h"
#include "dm_state.h"
#include "dm_defs.h"
#include "m_misc.h"
#include "m_swap.h"
#include "p_local.h"
#include "r_sky.h"
#include "w_wad.h"
#include "w_textur.h"
#include "z_zone.h"

// sprite lists
int numspritelumps;
int *spritelist = NULL;

// needed for pre rendering
float_t *spritewidth;
float_t *spriteheight;
float_t *spriteoffset;
float_t *spritetopoffset;

//
// InitFlats
//
static void InitFlats(void)
{
#if 0  // OLD STUFF, REMOVE SOON
  int i;
  int pos;

  if (!flatlist)
  {
    flatlist = W_GetList('F', &numflats);

#define CMP(a, b) (stricmp(W_GetLumpName(a), W_GetLumpName(b)) < 0)
    QSORT(int, flatlist, numflats, CUTOFF);
#undef CMP
  }
#endif
}

//
// R_AddFlatAnim
//
// Here are the rules, they get a bit hairy, but are the simplest
// thing which achieves expected behaviour.
//
// 1. When two flats in different wads have the same name, the flat
//    in the _later_ wad overrides the flat in the earlier wad.  This
//    allows pwads to replace iwad flats -- as is usual.
//
// 2. Thus the animation sequence is determined by the earliest wad
//    which contains _both_ the start and the end flat.
//
// 3. If the start and end flats are have different "newness" (e.g.
//    the start flat replaces one in the iwad, but the end flat is
//    new), the animation is _not_ valid.
//
// 4. Thus, defining a new animation sequence requires a new set of
//    flat names (different from those in the iwad).
//
// Rules 3/4 are a consequence of Rule 1 for replacement, as we can't
// assume that replaced flats will be in the same order as the
// originals, or that the intent was to create a new sequence with a
// (possibly) different set of flats in between.
// 
// -AJA- 2000/01/28: reworked flat animations.
// 
void R_AddFlatAnim(animdef_t *anim)
{
  int start = W_CheckNumForName(anim->startname);
  int end   = W_CheckNumForName(anim->endname);

  int s_file, e_file;
  int s_offset, e_offset;
  const int *lumps;
  
  if (start == -1 || end == -1)
  {
    // sequence not valid.  Maybe it is the DOOM 1 IWAD.
    return;
  }

  W_FindFlatInList(anim->startname, &s_file, &s_offset);
  W_FindFlatInList(anim->endname,   &e_file, &e_offset);

  if (s_file < 0)
  {
    I_Warning("Bad flat animation: `%s' not in any list.\n",
        anim->startname);
    return;
  }
  else if (e_file < 0)
  {
    I_Warning("Bad flat animation: `%s' not in any list.\n",
        anim->endname);
    return;
  }

  if (s_file != e_file)
  {
    I_Warning("Bad flat anim %s-%s: start and end in different wads.\n", 
        anim->startname, anim->endname);
    return;
  }

  if (s_offset > e_offset)
  {
    I_Warning("Bad flat anim: %s lump follows %s, should precede it.\n",
        anim->startname, anim->endname);
    return;
  }

  lumps = W_GetListLumps(s_file, LMPLST_Flats);
  
  W_AnimateFlatRange(lumps+s_offset, e_offset-s_offset+1, anim->speed);
}

//
// InitSpriteLumps
//
// Finds the width and hoffset of all sprites in the wad, so the sprite does
// not need to be cached completely just for having the header info ready
// during rendering.
//
static void InitSpriteLumps(void)
{
  int i;
  const patch_t *patch;

  if (!spritelist)
    spritelist = W_GetList('S', &numspritelumps);

  // -AJA- FIXME: clean this mess up
  spritewidth = Z_New(float_t, numspritelumps);
  spriteheight = Z_New(float_t, numspritelumps);
  spriteoffset = Z_New(float_t, numspritelumps);
  spritetopoffset = Z_New(float_t, numspritelumps);

  for (i = 0; i < numspritelumps; i++)
  {
    if (!(i & 63))
      I_Printf(".");

    patch = W_CacheLumpNum(spritelist[i]);

    spritewidth[i] = SHORT(patch->width);
    spriteheight[i] = SHORT(patch->height);
    spriteoffset[i] = SHORT(patch->leftoffset);
    spritetopoffset[i] = SHORT(patch->topoffset);

    W_DoneWithLump(patch);
  }
}

//
// R_InitData
//
// Locates all the lumps that will be used by all views.
// Must be called after W_Init.
//
void R_InitData(void)
{
  InitFlats();
  InitSpriteLumps();
}

//
// R_PrecacheLevel
//
// Preloads all relevant graphics for the level.
//
void R_PrecacheLevel(void)
{
// -AJA- needs reworking for image system, but is live-without-able.
#if 0
  char *flatpresent;
  char *texturepresent;
  char *spritepresent;
  int i;
  int j;
  int k;
  int numtextures;
  int lump;

  mobj_t *currmobj;

  spriteframe_t *sf;

  if (demoplayback)
    return;

  // Precache flats.
  flatpresent = Z_ClearNew(char, numflats);

  for (i = 0; i < numsectors; i++)
  {
    flatpresent[sectors[i].floor.pic] = 1;
    flatpresent[sectors[i].ceil.pic] = 1;
  }

  for (i = 0; i < numflats; i++)
  {
    if (flatpresent[i])
    {
      lump = flatlist[i];
      W_PreCacheLumpNum(lump);
    }
  }
  Z_Free(flatpresent);

  // Precache textures.
  texturepresent = NULL;
  numtextures = 0;

  // expands the texture array if needed, and marks the texture as used.
  // all used textures will be precached.
#define CHECK(num)\
if (num >= numtextures)\
{\
  Z_Resize(texturepresent, char, num + 1);\
  for (j = numtextures; j < num; j++)\
    texturepresent[j] = 0;\
  numtextures = num + 1;\
}\
texturepresent[num] = 1;

  for (i = 0; i < numsides; i++)
  {
    CHECK(sides[i].top.tex)
    CHECK(sides[i].middle.tex)
    CHECK(sides[i].bottom.tex)
  }
#undef CHECK

  //
  // Sky texture is always present.
  //
  // Note that F_SKY1 is the name used to indicate a sky floor/ceiling as a flat,
  // while the sky texture is stored like a wall texture, with a level depended name;
  // dictated to by the currentmap structure.
  //
  texturepresent[skytexture] = 1;

  // -AJA- EVIL: numtextures != sizeof(texturepresent) array
  for (i = 0; i < numtextures; i++)
  {
    if (!texturepresent[i])
      continue;

    W_PreCacheTextureNum(i);
  }
  Z_Free(texturepresent);

  // Precache sprites.
  spritepresent = Z_ClearNew(char, numsprites);

  for (currmobj = mobjlisthead; currmobj != NULL; currmobj = currmobj->next)
    spritepresent[currmobj->sprite] = 1;

  for (i = 0; i < numsprites; i++)
  {
    if (!spritepresent[i])
      continue;

    for (j = 0; j < sprites[i].numframes; j++)
    {
      sf = &sprites[i].spriteframes[j];

      for (k = 0; k < 8; k++)
      {
        lump = spritelist[sf->lump[k]];
        W_PreCacheLumpNum(lump);
      }
    }
  }
  Z_Free(spritepresent);
#endif  // ! USE_IMAGE
}
