//----------------------------------------------------------------------------
//  EDGE OpenGL Rendering (Unit batching)
//----------------------------------------------------------------------------
// 
//  Copyright (c) 1999-2000  The EDGE Team.
// 
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//----------------------------------------------------------------------------
//
// -AJA- 2000/10/09: Began work on this new unit system.
//

// this conditional applies to the whole file
#ifdef USE_GL

#include "i_defs.h"

#include "dm_defs.h"
#include "dm_state.h"
#include "e_search.h"
#include "m_bbox.h"
#include "m_random.h"
#include "p_local.h"
#include "p_mobj.h"
#include "r_defs.h"
#include "r_main.h"
#include "r_plane.h"
#include "r_sky.h"
#include "r_state.h"
#include "r_things.h"
#include "r2_defs.h"
#include "rgl_defs.h"
#include "v_colour.h"
#include "v_res.h"
#include "w_textur.h"
#include "w_wad.h"
#include "v_ctx.h"
#include "z_zone.h"


#define DEBUG  0


#define MAX_L_VERT  4096
#define MAX_L_UNIT  (MAX_L_VERT / 4)

// a single unit (polygon, quad, etc) to pass to the GL
typedef struct local_gl_unit_s
{
  // unit mode (e.g. GL_POLYGON)
  GLuint mode;

  // range of local vertices
  int first, count;
  
  // texture used
  GLuint tex_id;

  // texture contains see-through parts (i.e. significant areas where
  // alpha = 0, like sprites).
  boolean_t masked;

  // texture/vertices should be alpha blended (e.g. translucent
  // water).
  boolean_t blended;
}
local_gl_unit_t;

static local_gl_vert_t local_verts[MAX_L_VERT];
static local_gl_unit_t local_units[MAX_L_UNIT];

static GLuint local_unit_map[MAX_L_UNIT];

static int cur_vert;
static int cur_unit;

static boolean_t solid_mode;


//
// RGL_InitUnits
//
// Initialise the unit system.  Once-only call.
//
void RGL_InitUnits(void)
{
  // GL client state stuff disabled, due to problems with certain
  // drivers (in particular, Voodoo3 under Win32).
#if 0
  // setup pointers to client state
  glVertexPointer(3, GL_FLOAT, sizeof(local_gl_vert_t), &local_verts[0].x);
  glColorPointer (4, GL_FLOAT, sizeof(local_gl_vert_t), &local_verts[0].r);
  glTexCoordPointer(2, GL_FLOAT, sizeof(local_gl_vert_t), &local_verts[0].t_x);
  glNormalPointer(GL_FLOAT, sizeof(local_gl_vert_t), &local_verts[0].n_x);
  glEdgeFlagPointer(sizeof(local_gl_vert_t), &local_verts[0].edge);

  glEnableClientState(GL_EDGE_FLAG_ARRAY);
  glEnableClientState(GL_VERTEX_ARRAY);
  glEnableClientState(GL_COLOR_ARRAY);
  glEnableClientState(GL_TEXTURE_COORD_ARRAY);
  glEnableClientState(GL_NORMAL_ARRAY);
#endif
}

// RGL_StartUnits
//
// Starts a fresh batch of units.  There should be two batches of
// units, the first with solid == true (handling all solid
// walls/floors), and the second with solid == false (handling
// everything else: sprites, masked textures, translucent planes).
//
// The solid batch will be sorted to keep texture changes to a
// minimum.  The non-solid batch is drawn in-order (and should be
// processed from furthest to closest).
//
void RGL_StartUnits(boolean_t solid)
{
  cur_vert = cur_unit = 0;
  solid_mode = solid;
}

//
// RGL_FinishUnits
//
// Finishes a batch of units, drawing any that haven't been drawn yet.
//
void RGL_FinishUnits(void)
{
  RGL_DrawUnits();
}

//
// RGL_BeginUnit
//
// Begin a new unit, with the given parameters (mode and texture ID).
// `max_vert' is the maximum expected vertices of the quad/poly (the
// actual number can be less, but never more).  Returns a pointer to
// the first vertex structure.  `masked' should be true if the texture
// contains "holes" (like sprites).  `blended' should be true if the
// texture should be blended (like for translucent water or sprites).
//
local_gl_vert_t *RGL_BeginUnit(GLuint mode, int max_vert,
    GLuint tex_id, boolean_t masked, boolean_t blended)
{
  local_gl_unit_t *unit;
  
  DEV_ASSERT2(max_vert > 0);
  DEV_ASSERT2(tex_id != 0);

  // check for out-of-space
  if (cur_vert + max_vert > MAX_L_VERT || cur_unit >= MAX_L_UNIT)
  {
    RGL_DrawUnits();
  }

  unit = local_units + cur_unit;

  unit->mode    = mode;
  unit->tex_id  = tex_id;
  unit->first   = cur_vert;  // count set later
  unit->masked  = masked;
  unit->blended = blended;

  return local_verts + cur_vert;
}

//
// RGL_EndUnit
//
void RGL_EndUnit(int actual_vert)
{
  local_gl_unit_t *unit;

  DEV_ASSERT2(actual_vert > 0);

  unit = local_units + cur_unit;

  unit->count = actual_vert;

  cur_vert += actual_vert;
  cur_unit++;
  
  DEV_ASSERT2(cur_vert <= MAX_L_VERT);
  DEV_ASSERT2(cur_unit <= MAX_L_UNIT);
}

//
// RGL_DrawUnits
//
// Forces the set of current units to be drawn.  This call is
// optional (it never _needs_ to be called by client code).
//
void RGL_DrawUnits(void)
{
  int i, j;
  GLuint cur_tex = 0xABE74C74;
  GLfloat d_col[4] = { 1.0, 1.0, 1.0, 1.0 };

  boolean_t cur_masking  = false;
  boolean_t cur_blending = false;

  if (cur_unit == 0)
    return;
   
  for (i=0; i < cur_unit; i++)
    local_unit_map[i] = i;

  // need to sort ?
  if (solid_mode)
  {
    #define CMP(a,b)  \
        (local_units[a].tex_id < local_units[b].tex_id ||        \
         (local_units[a].tex_id == local_units[b].tex_id &&      \
          (local_units[a].blended < local_units[b].blended ||    \
           (local_units[a].blended == local_units[b].blended &&  \
            local_units[a].masked < local_units[b].masked))))
    QSORT(GLuint, local_unit_map, cur_unit, CUTOFF);
    #undef CMP
  }

  glEnable(GL_TEXTURE_2D);
  glDisable(GL_ALPHA_TEST);
  glDisable(GL_BLEND);

  for (i=0; i < cur_unit; i++)
  {
    local_gl_unit_t *unit = local_units + local_unit_map[i];
    
    DEV_ASSERT2(unit->count > 0);
    DEV_ASSERT2(unit->tex_id != 0);

    // detect changes in texture/alpha/blending and change state
    
    if (cur_masking != unit->masked)
    {
      cur_masking = unit->masked;

      if (cur_masking)
        glEnable(GL_ALPHA_TEST);
      else
        glDisable(GL_ALPHA_TEST);
    }

    if (cur_blending != unit->blended)
    {
      cur_blending = unit->blended;

      if (cur_blending)
        glEnable(GL_BLEND);
      else 
        glDisable(GL_BLEND);
    }

    if (cur_tex != unit->tex_id)
    {
      cur_tex = unit->tex_id;
      glBindTexture(GL_TEXTURE_2D, cur_tex);
    }

    // set alpha from first vertex (we never interpolate it)
    d_col[3] = local_verts[unit->first].col[3];
    glMaterialfv(GL_FRONT_AND_BACK, GL_DIFFUSE, d_col);

    glBegin(unit->mode);
 
    for (j=0; j < unit->count; j++)
    {
      local_gl_vert_t *V = local_verts + unit->first + j;

      glTexCoord2f(V->t_x, V->t_y);
      glNormal3f(V->n_x, V->n_y, V->n_z);
      glEdgeFlag(V->edge);
      glMaterialfv(GL_FRONT_AND_BACK, GL_EMISSION, V->col);

      // vertex must be last
      glVertex3f(V->x, V->y, V->z);
    }

    glEnd();
  }

  // all done
  cur_vert = cur_unit = 0;

  glDisable(GL_TEXTURE_2D);
  glDisable(GL_ALPHA_TEST);
  glDisable(GL_BLEND);
}


//----------------------------------------------------------------------------
//
//  SPECIAL 1D OCCLUSION BUFFER
//

#define ONED_POWER  12  // 4096 angles
#define ONED_SIZE   (1 << ONED_POWER)

// 1 bit per angle, packed into 32 bit values.
// (NOTE: for speed reasons, 1 is "clear", and 0 is "blocked")
//
static unsigned long oned_oculus_buffer[ONED_SIZE / 32];

// -AJA- these values could be computed (rather than looked-up)
// without too much trouble.  For now I want to get the logic correct.
//
static unsigned long oned_low_masks[32] =
{
  0xFFFFFFFF, 0x7FFFFFFF, 0x3FFFFFFF, 0x1FFFFFFF,
  0x0FFFFFFF, 0x07FFFFFF, 0x03FFFFFF, 0x01FFFFFF,
  0x00FFFFFF, 0x007FFFFF, 0x003FFFFF, 0x001FFFFF,
  0x000FFFFF, 0x0007FFFF, 0x0003FFFF, 0x0001FFFF,
  0x0000FFFF, 0x00007FFF, 0x00003FFF, 0x00001FFF,
  0x00000FFF, 0x000007FF, 0x000003FF, 0x000001FF,
  0x000000FF, 0x0000007F, 0x0000003F, 0x0000001F,
  0x0000000F, 0x00000007, 0x00000003, 0x00000001
};

static unsigned long oned_high_masks[32] =
{
  0x80000000, 0xC0000000, 0xE0000000, 0xF0000000,
  0xF8000000, 0xFC000000, 0xFE000000, 0xFF000000,
  0xFF800000, 0xFFC00000, 0xFFE00000, 0xFFF00000,
  0xFFF80000, 0xFFFC0000, 0xFFFE0000, 0xFFFF0000,
  0xFFFF8000, 0xFFFFC000, 0xFFFFE000, 0xFFFFF000,
  0xFFFFF800, 0xFFFFFC00, 0xFFFFFE00, 0xFFFFFF00,
  0xFFFFFF80, 0xFFFFFFC0, 0xFFFFFFE0, 0xFFFFFFF0,
  0xFFFFFFF8, 0xFFFFFFFC, 0xFFFFFFFE, 0xFFFFFFFF
};

#define LOW_MASK(L)   oned_low_masks[L]
#define HIGH_MASK(H)  oned_high_masks[H]


//
// RGL_1DOcclusionClear
//
// Clear all angles in the given range.  (Clear means open, i.e. not
// blocked).  The angles are relative to the VIEW angle, and must lie
// in the range -90 to +90 degrees (inclusive).
//
void RGL_1DOcclusionClear(angle_t low, angle_t high)
{
  unsigned int low_b, high_b;

  low  += ANG90;
  high += ANG90;

  DEV_ASSERT2(low  <= ANG180);
  DEV_ASSERT2(high <= ANG180);
  DEV_ASSERT2(low  <= high);

  if (low  == ANG180) low--;
  if (high == ANG180) high--;

  low  >>= (ANGLEBITS - ONED_POWER);
  high >>= (ANGLEBITS - ONED_POWER);

  low_b  = low  & 0x1F;  low  >>= 5; 
  high_b = high & 0x1F;  high >>= 5; 

  if (low == high)
  {
    oned_oculus_buffer[low] |= (LOW_MASK(low_b) & HIGH_MASK(high_b));
  }
  else
  {
    oned_oculus_buffer[low]  |= LOW_MASK(low_b);
    oned_oculus_buffer[high] |= HIGH_MASK(high_b);

    for (low++; low < high; low++)
      oned_oculus_buffer[low] = 0xFFFFFFFF;
  }
}

//
// RGL_1DOcclusionSet
//
// Set all angles in the given range, i.e. mark them as blocking.  The
// angles are relative to the VIEW angle, and must lie in the range
// -90 to +90 degrees (inclusive).
//
void RGL_1DOcclusionSet(angle_t low, angle_t high)
{
  unsigned int low_b, high_b;

  low  += ANG90;
  high += ANG90;

  DEV_ASSERT2(low  <= ANG180);
  DEV_ASSERT2(high <= ANG180);
  DEV_ASSERT2(low  <= high);

  if (low  == ANG180) low--;
  if (high == ANG180) high--;

  low  >>= (ANGLEBITS - ONED_POWER);
  high >>= (ANGLEBITS - ONED_POWER);

  low_b  = low  & 0x1F;  low  >>= 5; 
  high_b = high & 0x1F;  high >>= 5; 

  if (low == high)
  {
    oned_oculus_buffer[low] &= ~(LOW_MASK(low_b) & HIGH_MASK(high_b));
  }
  else
  {
    oned_oculus_buffer[low]  &= ~LOW_MASK(low_b);
    oned_oculus_buffer[high] &= ~HIGH_MASK(high_b);

    for (low++; low < high; low++)
      oned_oculus_buffer[low] = 0x00000000;
  }
}

//
// RGL_1DOcclusionTest
//
// Check whether all angles in the given range are set (i.e. blocked).
// Returns true if the entire range is blocked, false otherwise.
// Angles are relative to the VIEW angle, and must lie in the range
// -90 to +90 degrees (inclusive).
//
boolean_t RGL_1DOcclusionTest(angle_t low, angle_t high)
{
  unsigned int low_b, high_b;

  low  += ANG90;
  high += ANG90;

  DEV_ASSERT2(low  <= ANG180);
  DEV_ASSERT2(high <= ANG180);
  DEV_ASSERT2(low  <= high);

  if (low  == ANG180) low--;
  if (high == ANG180) high--;

  low  >>= (ANGLEBITS - ONED_POWER);
  high >>= (ANGLEBITS - ONED_POWER);

  // make the check a little wider, to be sure to be sure
  if (low > 0) low--;
  if (high < (ONED_SIZE-1)) high++;
 
  low_b  = low  & 0x1F;  low  >>= 5; 
  high_b = high & 0x1F;  high >>= 5; 

  if (low == high)
    return (oned_oculus_buffer[low] & 
        LOW_MASK(low_b) & HIGH_MASK(high_b)) ? false : true;

  if (oned_oculus_buffer[low] & LOW_MASK(low_b))
    return false;
   
  if (oned_oculus_buffer[high] & HIGH_MASK(high_b))
    return false;

  for (low++; low < high; low++)
    if (oned_oculus_buffer[low])
      return false;

  return true;
}

#endif  // USE_GL
