//----------------------------------------------------------------------------
//  EDGE New SaveGame Handling (Writing)
//----------------------------------------------------------------------------
// 
//  Copyright (c) 1999-2001  The EDGE Team.
// 
//  This program is free software; you can redistribute it and/or
//  modify it under the terms of the GNU General Public License
//  as published by the Free Software Foundation; either version 2
//  of the License, or (at your option) any later version.
//
//  This program is distributed in the hope that it will be useful,
//  but WITHOUT ANY WARRANTY; without even the implied warranty of
//  MERCHANTABILITY or FITNESS FOR A PARTICULAR PURPOSE.  See the
//  GNU General Public License for more details.
//
//----------------------------------------------------------------------------
//
// See the file "docs/save_sys.txt" for a complete description of the
// new savegame system.
//

#include "i_defs.h"
#include "sv_chunk.h"

#include "dm_state.h"
#include "e_main.h"
#include "g_game.h"
#include "m_math.h"
#include "m_misc.h"
#include "m_random.h"
#include "p_local.h"
#include "p_spec.h"
#include "r_state.h"
#include "sv_chunk.h"
#include "sv_main.h"
#include "w_wad.h"
#include "wi_stuff.h"
#include "z_zone.h"


//
// SV_BeginSave
//
void SV_BeginSave(void)
{
  /* does nothing */
}

//
// SV_FinishSave
//
void SV_FinishSave(void)
{
  /* does nothing */
}

//
// SV_SaveStruct
//
void SV_SaveStruct(void *base, savestruct_t *info)
{
  savefield_t *cur;
  char *storage;
  int i;
  
  SV_PushWriteChunk(info->marker);
  
  for (cur=info->fields; cur->type.kind != SFKIND_Invalid; cur++)
  {
    // ignore read-only (fudging) fields
    if (! cur->field_put)
      continue;
    
    storage = ((char *)base) + cur->offset;

    for (i=0; i < cur->count; i++)
    {
      switch (cur->type.kind)
      {
        case SFKIND_Struct:
        case SFKIND_Index:
          (* cur->field_put)(storage, i, (char*)cur->type.name);
          break;
        
        default:
          (* cur->field_put)(storage, i, NULL);
          break;
      }
    }
  }

  SV_PopWriteChunk();
}

static void GlobWriteWHER(void)
{
  // first is the [WHER] chunk
  SV_PushWriteChunk("Wher");
  SV_PutString(currentmap->interscr->ddf.name);  // game name
  SV_PutString(currentmap->ddf.name);  // level name
  SV_PutInt(G_GetLevelFlags());        // level flags
  SV_PutInt(level_flags.menu_grav);    // gravity
  SV_PopWriteChunk();
}

static void GlobWriteWHEN(void)
{
  // second is the [WHEN] chunk
  SV_PushWriteChunk("When");
  SV_PutInt(leveltime);             // level time
  SV_PutInt(P_ReadRandomState());   // random snapshot
  SV_PutInt(totalkills);            // total kills
  SV_PutInt(totalitems);            // total items
  SV_PutInt(totalsecret);           // total secrets
  SV_PopWriteChunk();
}

static void GlobWriteSKIL(void)
{
  // third is the [SKIL] chunk
  SV_PushWriteChunk("Skil");
  SV_PutInt(consoleplayer->pnum);   // console player
  SV_PutInt(gameskill);             // skill level
  SV_PutInt(netgame ? (1+deathmatch) : 0);  // netgame mode
  SV_PopWriteChunk();
}

static void GlobWriteDESC(const char *description)
{
  time_t cur_time;
  char timebuf[100];
  
  // fourth is the [DESC] chunk
  time(&cur_time);
  strftime(timebuf, 99, "%I:%M %p  %d/%b/%Y", localtime(&cur_time));
 
  SV_PushWriteChunk("Desc");
  SV_PutString(description);    // description
  SV_PutString(timebuf);        // date & time
  SV_PopWriteChunk();
}

static void GlobWriteVIEW(void)
{
  int x, y;
  
//!!!!  if (! save_screenshot_valid)
//!!!!    return;
  
  SV_PushWriteChunk("View");
  SV_PutInt(160);             // width
  SV_PutInt(100);             // height
  
  for (y=0; y < 100; y++)
  for (x=0; x < 160; x++)
  {
    SV_PutShort(save_screenshot[x][y]);
  }

  SV_PopWriteChunk();
}

static void GlobWriteWDCK(void)
{
  //!!! IMPLEMENT THIS
}

static void GlobWriteLVCK(void)
{
  //!!! IMPLEMENT THIS
}

static void GlobWriteDFCK(void)
{
  //!!! IMPLEMENT THIS
}

//
// SV_SaveGLOB
//
void SV_SaveGLOB(const char *description)
{
  SV_PushWriteChunk("Glob");

  GlobWriteWHER();
  GlobWriteWHEN();
  GlobWriteSKIL();

  GlobWriteDESC(description);
  GlobWriteVIEW();

  GlobWriteWDCK();
  GlobWriteLVCK();
  GlobWriteDFCK();

  // all done with [GLOB]
  SV_PopWriteChunk();
}

static void SV_SaveSDEF(savestruct_t *S)
{
  int i, num;
  savefield_t *F;

  // count number of fields
  for (num=0; S->fields[num].type.kind != SFKIND_Invalid; num++)
  { /* nothing here */ }

  SV_PutInt(num);

  SV_PutString(S->struct_name);
  SV_PutString(S->marker);

  // write out the fields

  for (i=0, F=S->fields; i < num; i++, F++)
  {
    SV_PutByte((unsigned char) F->type.kind);
    SV_PutByte((unsigned char) F->type.size);
    SV_PutShort((unsigned short) F->count);
    SV_PutString(F->field_name);

    if (F->type.kind == SFKIND_Struct ||
        F->type.kind == SFKIND_Index)
    {
      SV_PutString(F->type.name);
    }
  }
}

//
// SV_SaveSTRU
//
void SV_SaveSTRU(void)
{
  savestruct_t *cur;
  
  SV_PushWriteChunk("Stru");

  for (cur=sv_known_structs; cur; cur=cur->next)
  {
    if (! cur->define_me)
      continue;

    SV_PushWriteChunk("Sdef");
    SV_SaveSDEF(cur);
    SV_PopWriteChunk();
  }

  SV_PopWriteChunk();
}


static void SV_SaveADEF(savearray_t *A)
{
  int num_elem = (* A->count_elems)();

  SV_PutInt(num_elem);

  SV_PutString(A->array_name);
  SV_PutString(A->sdef->struct_name);
}

//
// SV_SaveARRY
//
void SV_SaveARRY(void)
{
  savearray_t *cur;
  
  SV_PushWriteChunk("Arry");

  for (cur=sv_known_arrays; cur; cur=cur->next)
  {
    if (! cur->define_me)
      continue;

    SV_PushWriteChunk("Adef");
    SV_SaveADEF(cur);
    SV_PopWriteChunk();
  }

  SV_PopWriteChunk();
}

static void SV_SaveSTOR(savearray_t *A)
{
  int num_elem = (* A->count_elems)();
  int i;
  
  SV_PutString(A->array_name);

  for (i=0; i < num_elem; i++)
  {
    sv_current_elem = (* A->get_elem)(i);

    DEV_ASSERT2(sv_current_elem);

    SV_SaveStruct(sv_current_elem, A->sdef);
  }
}

//
// SV_SaveDATA
//
void SV_SaveDATA(void)
{
  savearray_t *cur;
  
  for (cur=sv_known_arrays; cur; cur=cur->next)
  {
    if (! cur->define_me)
      continue;

    SV_PushWriteChunk("Stor");
    SV_SaveSTOR(cur);
    SV_PopWriteChunk();
  }
}

